############################################
#
#  auxilliary functions needed for the JGL.plug function
#
#############################################

admm.iters.unconnected.plug <- function (S.list, lambda1, lambda2, penalty = "fused", rho = 1, rho.increment = 1, 
                                          weights, maxit = 500, tol = 1e-05, warm = NULL) 
{
  # this function is an adaptation of the admm.iters.unconnected function from the JGL package
  #########
  # INPUTS #
  #########
  # S.list : list of estimated variances of unconnected features
  # penalty : penalty of JGL (possible values "fused" and "group")
  # lambda1 : regularization parameter on the elements of inverse error covariance matrix
  # lambda2 : regularization parameter on the difference between elements of inverse error covariance matrix (fused lasso)
  # rho : step size parameter : large value decrease step size
  # weights : putative sample size of each class'data . Allowed values : a vector of length K; 
  # penalize.diagonal :determines wether the penalties are applied to the diagonal elements
  # maxit : maximum number of iterations
  # tol : convergence criterion
  # warm :  input a warm start to theta in the form of a K-length list of pxp matrices

  ##########
  # OUTPUTS #
  ##########
  # theta : list of the estimated inverse covariance matrices
  # Z : new variables of the ADMM algorithm 
  # diff: matrix of sum of the K differences between theta_k and Z_k
  # iters: number of iterations
  
  n = weights
  S = S.list
  p = length(S[[1]])
  K = length(S)
  
  # initialize theta
    theta = vector("list",K)
    for (k in 1:K) {
      theta[[k]] = 1/S[[k]]
    }
  # initialize Z
    Z = vector("list",K)
    for (k in 1:K) {
      Z[[k]] = rep(0, p)
    }
  # initialize the dual varaibles W
    W = vector("list",K)
    for (k in 1:K) {
      W[[k]] = rep(0, p)
    }
  # initialize lambdas
    lam1 = lambda1
    lam2 = lambda2
    
  # iterations
    iter = 0
    diff_value = 10
    while ((iter == 0) || (iter < maxit && diff_value > tol)) {
      
      # update theta
      theta.prev = theta
      for (k in 1:K) {
        B = n[k] * S[[k]] - rho * (Z[[k]] - W[[k]])
        theta[[k]] = 1/(2 * rho) * (-B + sqrt(B^2 + 4 * rho * n[k]))
      }
     # update Z   
      A = list()
      for (k in 1:K) {
        A[[k]] = theta[[k]] + W[[k]]
      }
      if (penalty == "fused") {
        # use flsa to minimize rho/2 ||Z-A||_F^2 + P(Z):
        if (K == 2) { Z = flsa2(A, rho, lam1, lam2, penalize.diagonal = TRUE) }
        if (K > 2) { Z = flsa.general(A, rho, lam1, lam2, penalize.diagonal = TRUE)}
      }
      if (penalty == "group") { 
        # minimize rho/2 ||Z-A||_F^2 + P(Z):
        Z = dsgl(A, rho, lam1, lam2, penalize.diagonal = TRUE)
      }
    # update W
      for (k in 1:K) {
        W[[k]] = W[[k]] + (theta[[k]] - Z[[k]])
      }
    # book keeping 
    iter = iter + 1
    diff_value = 0
    for (k in 1:K) {
      diff_value = diff_value + sum(abs(theta[[k]] - theta.prev[[k]]))/sum(abs(theta.prev[[k]]))
    }
    #increment rho by a constant factor
    rho = rho * rho.increment
  }
  diff = 0
  for (k in 1:K) {
    diff = diff + sum(abs(theta[[k]] - Z[[k]]))
  }
  out = list(theta = theta, Z = Z, diff = diff, iters = iter)
  return(out)
}



#################################
admm.iters.plug <- function (S.list, lambda1, lambda2, 
                             penalty = "fused", rho = 1, rho.increment = 1, 
                              weights="sample size", penalize.diagonal, maxit = 500, tol = 10e-05, 
                              warm = NULL) 
{
  # This function is an adaptation of the admm.iters function from the JGL package
  #########
  # INPUT #
  #########
  # S.list : list of estimated covariance matrices for JGL
  # penalty : penalty of JGL (possible values "fused" and "group")
  # lambda1 : regularization parameter on the elements of inverse error covariance matrix
  # lambda2 : regularization parameter on the difference between elements of inverse error covariance matrix (fused lasso)
  # rho : step size parameter : large value decrease step size
  # weights : 
  # penalize.diagonal :determines wether the penalties are applied to the diagonal elements
  # maxit : maximum number of iterations
  # tol : convergenec criterion
  # warm :  input a warm start to theta in the form of a K-length list of pxp matrices
  ##########
  # OUTPUT #
  ##########
  # theta : list of the estimated inverse covariance matrices
  # Z : new variables of the ADMM algorithm 
  # diff: matrix of sum of the K differences between theta_k and Z_k
  # iters: number of iterations
  
  n = weights
  S = S.list
  p = dim(S.list[[1]])[2]
  K=length(S)
  
  # initialize theta (diagonal matrix with inverse variances as diagonal)
  theta = vector("list",K)
  for (k in 1:K) {
    theta[[k]] = diag(1/diag(S[[k]]))
  }
  # initialize Z
  Z = vector("list",K)
  for (k in 1:K) {
    Z[[k]] = matrix(0, p, p)
  }
  # initialize W (dual variables)
  W = vector("list",K)
  for (k in 1:K) {
    W[[k]] = matrix(0, p, p)
  }
  
  # initialize lambdas
  lam1 = penalty.as.matrix(lambda1, p, penalize.diagonal = penalize.diagonal)
  if (penalty == "fused") {
    lam2 = penalty.as.matrix(lambda2, p, penalize.diagonal = TRUE)
  }
  if (penalty == "group") {
    lam2 = penalty.as.matrix(lambda2, p, penalize.diagonal = penalize.diagonal)
  }
  
  
  # iterations
  iter = 0
  diff_value = 10
  
  while ((iter == 0) || (iter < maxit && diff_value > tol)) {
    
    # only for reporting
      if (FALSE) {
        print(paste("iter=", iter))
        if (penalty == "fused") {
          print(paste("crit=", crit(theta, S, n = rep(1, K), lam1, lam2, penalize.diagonal = penalize.diagonal)))
          print(paste("crit=", crit(Z, S, n = rep(1, K),lam1, lam2, penalize.diagonal = penalize.diagonal)))
        }
        if (penalty == "group") {
          print(paste("crit=", gcrit(theta, S, n = rep(1, K), lam1, lam2, penalize.diagonal = penalize.diagonal)))
        }
      }
    
    # update theta
      theta.prev = theta
      for (k in 1:K) {
        edecomp = eigen(S[[k]] - rho * Z[[k]]/n[k] + rho * W[[k]]/n[k])
        D = edecomp$values
        V = edecomp$vectors
        D2 = n[k]/(2 * rho) * (-D + sqrt(D^2 + 4 * rho/n[k]))
        theta[[k]] = V %*% diag(D2) %*% t(V)
      }
    
    # update Z
      A = list()
      for (k in 1:K) {
        A[[k]] = theta[[k]] + W[[k]]
      }
      if (penalty == "fused") {
        # use flsa (fused lasso signal approximation) to minimize rho/2 ||Z-A||_F^2 + P(Z)
        if (K == 2) {  Z = flsa2(A, rho, lam1, lam2, penalize.diagonal = TRUE) } # simple closed form solution
        if (K > 2) {  Z = flsa.general(A, rho, lam1, lam2, penalize.diagonal = TRUE)}
      }
      
      if (penalty == "group") {
        # minimize rho/2 ||Z-A||_F^2 + P(Z) : simpel closed form solution
        Z = dsgl(A, rho, lam1, lam2, penalize.diagonal = TRUE)
      }
      
    
    # update W
      for (k in 1:K) {
        W[[k]] = W[[k]] + (theta[[k]] - Z[[k]])
      }
      
    # bookeeping  
      iter = iter + 1
      diff_value = 0
      for (k in 1:K) {
        diff_value = diff_value + sum(abs(theta[[k]] - theta.prev[[k]]))/sum(abs(theta.prev[[k]]))
      }
      rho = rho * rho.increment # increment rho by a constant factor
  }
  
  diff = 0
  for (k in 1:K) {
    diff = diff + sum(abs(theta[[k]] - Z[[k]]))
  }
  out = list(theta = theta, Z = Z, diff = diff, iters = iter)
  return(out)
}

flsa2<-function (A, L, lam1, lam2, penalize.diagonal) 
{
  # function from the JGL package
  S1 = abs(A[[1]] - A[[2]]) <= 2 * lam2/L
  X1 = (A[[1]] + A[[2]])/2
  Y1 = X1
  S2 = (A[[1]] > A[[2]] + 2 * lam2/L)
  X2 = A[[1]] - lam2/L
  Y2 = A[[2]] + lam2/L
  S3 = (A[[2]] > A[[1]] + 2 * lam2/L)
  X3 = A[[1]] + lam2/L
  Y3 = A[[2]] - lam2/L
  X = soft(a = S1 * X1 + S2 * X2 + S3 * X3, lam = lam1/L, penalize.diagonal = penalize.diagonal)
  Y = soft(a = S1 * Y1 + S2 * Y2 + S3 * Y3, lam = lam1/L, penalize.diagonal = penalize.diagonal)
  return(list(X, Y))
}



flsa.general<-function (A, L, lam1, lam2, penalize.diagonal) 
{
  # function from the JGL package
  trueA = A
  if (is.matrix(A[[1]])) {
    p = dim(A[[1]])[1]
  }
  if (is.vector(A[[1]])) {
    p = length(A[[1]])
  }
  K = length(A)
  X = list()
  for (k in 1:K) {
    X[[k]] = A[[1]] * NA
  }
  if (is.matrix(A[[1]])) {
    fusions = array(FALSE, dim = c(K, K, p, p))
  }
  if (is.vector(A[[1]])) {
    fusions = array(FALSE, dim = c(K, K, p, 1))
  }
  newc = list()
  for (k in 1:K) {
    others = setdiff(1:K, k)
    others.smaller.k = 1:(k - 1)
    newc[[k]] = A[[1]] * 0
    for (o in others) {
      newc[[k]] = newc[[k]] + (A[[o]] - A[[k]] < -1e-04) - 
        (A[[o]] - A[[k]] > 1e-04)
    }
  }
  for (iter in 1:(K - 1)) {
    ordermats = vector("list",K)
    for (k in 1:K) {
      others = setdiff(1:K, k)
      others.smaller.k = 1:(k - 1)
      ordermats[[k]] = A[[1]] * 0
      for (o in others) {
        ordermats[[k]] = ordermats[[k]] + (A[[k]] - A[[o]] > 
                                             1e-04)
      }
      if (k > 1) {
        for (o in others.smaller.k) {
          ordermats[[k]] = ordermats[[k]] + (abs(A[[o]] - 
                                                   A[[k]]) < 1e-04)
        }
      }
      ordermats[[k]] = ordermats[[k]] + 1
    }
    betas.g = list()
    for (k in 1:K) {
      betas.g[[k]] = A[[k]] - lam2/L * newc[[k]]
    }
    new.ordermats = list()
    for (k in 1:K) {
      others = setdiff(1:K, k)
      others.smaller.k = 1:(k - 1)
      new.ordermats[[k]] = A[[1]] * 0
      for (o in others) {
        new.ordermats[[k]] = new.ordermats[[k]] + (betas.g[[k]] - 
                                                     betas.g[[o]] > 1e-04)
      }
      if (k > 1) {
        for (o in others.smaller.k) {
          new.ordermats[[k]] = new.ordermats[[k]] + (abs(betas.g[[o]] - 
                                                           betas.g[[k]]) < 1e-04)
        }
      }
      new.ordermats[[k]] = new.ordermats[[k]] + 1
    }
    for (k in 1:K) {
      for (kp in 1:K) {
        fusions[k, kp, , ] = fusions[k, kp, , ] + ((ordermats[[k]] - 
                                                      1 == ordermats[[kp]]) & (new.ordermats[[k]] < 
                                                                                 new.ordermats[[kp]])) + ((ordermats[[k]] + 
                                                                                                             1 == ordermats[[kp]]) & (new.ordermats[[k]] > 
                                                                                                                                        new.ordermats[[kp]])) + (abs(A[[k]] - A[[kp]]) < 
                                                                                                                                                                   1e-04)
        fusions = (fusions > 0) * 1
      }
    }
    for (k in 1:K) {
      for (kp in 1:K) {
        others = setdiff(1:K, c(k, kp))
        for (o in others) {
          bothfused = fusions[k, o, , ] & fusions[kp, 
                                                  o, , ]
          fusions[k, kp, , ] = fusions[k, kp, , ] | bothfused
        }
      }
    }
    for (k in 1:K) {
      others = setdiff(1:K, k)
      fusemean = trueA[[k]]
      denom = A[[1]] * 0 + 1
      for (o in others) {
        fusemean = fusemean + fusions[k, o, , ] * trueA[[o]]
        denom = denom + fusions[k, o, , ]
      }
      A[[k]] = fusemean/denom
    }
    newc = list()
    for (k in 1:K) {
      others = setdiff(1:K, k)
      others.smaller.k = 1:(k - 1)
      newc[[k]] = A[[1]] * 0
      for (o in others) {
        newc[[k]] = newc[[k]] + (A[[o]] - A[[k]] < -1e-04) - 
          (A[[o]] - A[[k]] > 1e-04)
      }
    }
  }
  for (k in 1:K) {
    betas.g[[k]] = A[[k]] - lam2/L * newc[[k]]
  }
  for (k in 1:K) {
    X[[k]] = soft(betas.g[[k]], lam = lam1/L, penalize.diagonal = penalize.diagonal)
  }
  return(X)
}

#getAnywhere(soft)
soft<-function (a, lam, penalize.diagonal) 
{
  out <- sign(a) * pmax(0, abs(a) - lam)
  if (!penalize.diagonal) 
    diag(out) <- diag(a)
  return(out)
}


penalty.as.matrix<-function (lambda, p, penalize.diagonal) 
{
  # function from the JGL package
  if (is.matrix(lambda)) {
    if (sum(lambda != t(lambda)) > 0) {
      stop("error: penalty matrix is not symmetric")
    }
    if (sum(abs(dim(lambda) - p)) != 0) {
      stop("error: penalty matrix has wrong dimension")
    }
  }
  if (length(lambda) == 1) {
    lambda = matrix(lambda, p, p)
  }
  if (!penalize.diagonal) {
    diag(lambda) = 0
  }
  return(lambda)
}
