#############
# GL-LDA
#############

GLLDA<-function(trainY, groupings, testY,
                        estim="class",
                        Maxmin.ratio=0.01,  nsteps=5){
  
  ######################
  # INPUTS
  ######################
  # trainY : Data matrix of dimension N X p (N: total sample size, p: number of variables)
  # groupings : numeric vector of dimension N containing the group membership of the observations
  # testY : Data matrix of dimension m X p  containing the m new observations to classify
  # estim : character string for the choice of location and covariance estimates
  #         either "class" for the sample estimators, or "pairwCorrK" for the cordinatewise median and the pairwise covariance matrices
  #         based on Kendall's correlation
  # Maxmin.ratio : ratio between the lower bound and the upper bound of the grid of values for the parameter lambda1
  # nsteps: number of parameter values in the grid search for the optimal lambda1
  
  ######################
  # OUTPUTS
  ######################
  # estim: choice of location and covariance estimates 
  # mean_list : list of mean estimates for each group
  # S.list : list of covariance matrix estimates for each group
  # S.pooled : pooled covariance matrix
  # theta: estimated pooled precision matrix (output of the GLasso)
  # lambda1_max: maximum value of the grid of values for lambda1
  # lambda1_opt : optimal regularization parameter minimizing the BIC
  # classes: predicted groups for the observation of testY
  
  ######################
  # LIBRARIES
  ######################
  # library(huge)
  # library(pcaPP)
  # library(robustbase)
  
  ######################
  # AUXILIARY FUNCTIONS
  ######################
  # CellwiseCov.R
  # BICfct.R
  # predictDA.R
  
  ######################
  # ADDITIONAL FUNCTIONS
  ######################
  
  ##########
  # START
  ##########
  
  trainY<-split(data.frame(trainY),groupings)
  K<-length(unique(groupings))
  n<-unlist(lapply(trainY,nrow))
  p<-dim(trainY[[1]])[2]
  
  # Estimation of the means and covariance matrices
  if(estim=="class"){
    mean_list <- lapply(trainY, function(u)apply(u,2,mean))
    S.list<-lapply(trainY, cov)
  }else if(estim=="pairwCorrK"){
    mean_list<- lapply(trainY,function(t) apply(t,2,median))
    S.list<- lapply(trainY, kendall.transformed)
  }
  n<-unlist(lapply(trainY,nrow))
  S.pooled<- Reduce('+',mapply(function(s,n)n*s,s=S.list,n=n,SIMPLIFY=F))/(sum(n)-K)
  
  # Grid of regularization parameters
  l.gllda_max<- max(max(S.pooled - diag(p)), -min(S.pooled - diag(p)))
  l.gllda<-exp(seq(log(l.gllda_max),log(Maxmin.ratio*l.gllda_max),length.out=nsteps))
  
  # Selection of the regularization parameters
  FIT<- huge(S.pooled, lambda=l.gllda, method="glasso",verbose=F)$icov
  BICval.GLLDA<- unlist(lapply( FIT, function(t) BICfct(A=S.list,B=replicate(K,t),n=n,p=p)))
  l.gllda_opt.BIC<- l.gllda[which.min(BICval.GLLDA)]

  # Final estimation with optimal parameters
  GL_LDA.BIC<- huge(S.pooled,method="glasso",verbose=F,lambda=l.gllda_opt.BIC)$icov
  classes<-predict.DA(testY,means=mean_list, theta=GL_LDA.BIC[[1]], type="LDA",priors=n/sum(n))$cl
  
  return(list( estim=estim,mean_list=mean_list,S.list=S.list, S.pooled=S.pooled,
               theta= GL_LDA.BIC, 
               lambda1_max=l.gllda_max,lambda1_opt=l.gllda_opt.BIC,
               predict = classes))
  
}


  