#############
# GL-QDA
#############

GLQDA<-function(trainY, groupings, testY,
                estim="class",
                Maxmin.ratio=0.01,  nsteps=5){
  
  ######################
  # INPUTS
  ######################
  # trainY : Data matrix of dimension N X p (N: total sample size, p: number of variables)
  # groupings : numeric vector of dimension N containing the group membership of the observations
  # testY : Data matrix of dimension m X p  containing the m new observations to classify
  # estim : character string for the choice of location and covariance estimates
  #         either "class" for the sample estimators, or "pairwCorrK" for the cordinatewise median and the pairwise covariance matrices
  #         based on Kendall's correlation
  # Maxmin.ratio : ratio between the lower bound and the upper bound of the grid of values for the parameter lambda1
  # nsteps: number of parameter values in the grid search for the optimal lambda1
  
  ######################
  # OUTPUTS
  ######################
  # estim: choice of location and covariance estimates 
  # mean_list : list of mean estimates for each group
  # S.list : list of covariance matrix estimates for each group
  # theta: estimated precision matrices (output of the GLasso in each group)
  # lambda1_max: maximum value of the grid of values for lambda1
  # lambda1_opt : optimal regularization parameter minimizing the BIC
  # classes: predicted groups for the observation of testY
  
  ######################
  # LIBRARIES
  ######################
  # library(huge)
  # library(pcaPP)
  # library(robustbase)
  
  ######################
  # AUXILIARY FUNCTIONS
  ######################
  # CellwiseCov.R
  # BICfct.R
  # predictDA.R
  
  ######################
  # ADDITIONAL FUNCTIONS
  ######################
  
  BIC.GL<-function(lambda, S.list, n, p){
    FIT<- lapply(S.list, huge, lambda=lambda, method="glasso",verbose=F)
    BIC_values<-  BICfct( A= S.list, B=sapply(FIT,with,icov), n=n,p=p)
    return(BIC=BIC_values)
  }
  
  ##########
  # START
  ##########
  
  trainY<-split(data.frame(trainY),groupings)
  K<-length(unique(groupings))
  n<-unlist(lapply(trainY,nrow))
  p<-dim(trainY[[1]])[2]
  
  # Estimation of the means and covariance matrices
  if(estim=="class"){
    mean_list <- lapply(trainY, function(u)apply(u,2,mean))
    S.list<-lapply(trainY, cov)
  }else if(estim=="pairwCorrK"){
    mean_list<- lapply(trainY,function(t) apply(t,2,median))
    S.list<- lapply(trainY, kendall.transformed)
  }
  S.pooled<- Reduce('+',mapply(function(s,n)n*s,s=S.list,n=n,SIMPLIFY=F))/(sum(n)-K)
  
  # Grid of regularization parameters
  l.glqda_max<-max(unlist(lapply(S.list, function(s) max(max(s-diag(p)),-min(s-diag(p))))))
  l.glqda<-exp(seq(log(l.glqda_max),log(Maxmin.ratio*l.glqda_max),length.out=nsteps))
  
  # Selection of the regularization parameters
  BICval.GLQDA<- lapply(l.glqda, BIC.GL, S.list=S.list,n=n,p=p) 
  l.glqda_opt.BIC<- l.glqda[which.min(BICval.GLQDA)]

  #Final estimation with optimal parameters
  GL_QDA.BIC<- mapply(function(x,lambda)huge(x,lambda,method="glasso", verbose="F")$icov,x=S.list, lambda=l.glqda_opt.BIC )
  classes<-predict.DA(testY,means=mean_list, theta=GL_QDA.BIC, type="QDA",priors=n/sum(n))$cl
  
  return(list( estim=estim,mean_list=mean_list,S.list=S.list, 
               theta= GL_QDA.BIC, 
               lambda1_max=l.glqda_max, lambda1_opt=l.glqda_opt.BIC,
               predict = classes))
}