#############
# JGL-DA
#############

JGLDA<-function(trainY, groupings, testY,
                estim="class",
                Maxmin.ratio=0.01, Maxmin.ratio2=0.01, nsteps=5,nsteps2=5){
  
  
  ######################
  # INPUTS
  ######################
  # trainY : Data matrix of dimension N X p (N: total sample size, p: number of variables)
  # groupings : numeric vector of dimension N containing the group membership of the observations
  # testY : Data matrix of dimension m X p  containing the m new observations to classify
  # estim : character string for the choice of location and covariance estimates
  #         either "class" for the sample estimators, or "pairwCorrK" for the cordinatewise median and the pairwise covariance matrices
  #         based on Kendall's correlation
  # Maxmin.ratio : ratio between the lower bound and the upper bound of the grid of values for the parameter lambda1
  # Maxmin.ratio2 : ratio between the lower bound and the upper bound of the grid of values for the parameter lambda2
  # nsteps: number of parameter values in the grid search for the optimal lambda1
  # nsteps2: number of parameter values in the grid search for the optimal lambda2
  
  ######################
  # OUTPUTS
  ######################
  # estim: choice of location and covariance estimates 
  # mean_list : list of mean estimates for each group
  # S.list : list of covariance matrix estimates for each group
  # theta: estimated precision matrices (output of the JGL)
  # lambda1_max: maximum value of the grid of values for lambda1
  # lambda1_opt : optimal regularization parameter lambda1 minimizing the BIC
  # lambda1_max: maximum value of the grid of values for lambda2
  # lambda1_opt : optimal regularization parameter lambda2 minimizing the BIC
  # classes: predicted groups for the observation of testY
  
  ######################
  # LIBRARIES
  ######################
  # library(huge)
  # library(pcaPP)
  # library(robustbase)
  
  ######################
  # AUXILIARY FUNCTIONS
  ######################
  # CellwiseCov.R
  # BICfct.R
  # predictDA.R
  # JGL_plug.R
  # ADMM_plug.R
  
  ######################
  # ADDITIONAL FUNCTIONS
  ######################
  BIC.jgl <-function( lambda1, lambda2, S.list , penalty, n, p){
    FIT<-JGL.plug(S.list=S.list,n=n,  lambda1=lambda1, lambda2=lambda2) 
    BIC_value<- BICfct(A=S.list, B= FIT$theta, n=n,p=p)
    return(BIC_value)
  }
  ##########
  # START
  ##########
  
  trainY<-split(data.frame(trainY),groupings)
  K<-length(unique(groupings))
  n<-unlist(lapply(trainY,nrow))
  p<-dim(trainY[[1]])[2]
  
  # Estimation of the means and covariance matrices
  if(estim=="class"){
    mean_list <- lapply(trainY, function(u)apply(u,2,mean))
    S.list<-lapply(trainY, cov)
  }else if(estim=="pairwCorrK"){
    mean_list<- lapply(trainY,function(t) apply(t,2,median))
    S.list<- lapply(trainY, kendall.transformed)
  }
  S.pooled<- Reduce('+',mapply(function(s,n)n*s,s=S.list,n=n,SIMPLIFY=F))/(sum(n)-K)
  
  # Grid of regularization paramaters
  lambda1_max<- max(unlist(mapply(function(n,s) max(n*abs(s-diag(diag(s)))),n=n,s=S.list))) 
  lambda1_seq<-exp(seq(log(lambda1_max),log(Maxmin.ratio*lambda1_max),length.out=nsteps))
  lambda2_max<- max(unlist(mapply(function(s,n) n*max(abs(S.pooled-s)), n=n,s=S.list)))
  lambda2_seq<-exp(seq(log(lambda2_max),log(Maxmin.ratio2*lambda2_max),length.out=nsteps2))
  l.jgl<-list(lambda1_l=sort(rep(lambda1_seq,length(lambda2_seq)),decreasing=T), lambda2_l=rep(lambda2_seq,length(lambda1_seq)))

  # Selection of the regularization parameters
  morearg.plug=list(S.list=S.list,penalty="fused", n=n, p=p)
  BICval.JGL<- mapply(BIC.jgl,l.jgl$lambda1_l, l.jgl$lambda2_l, MoreArgs = morearg.plug,SIMPLIFY=F)
  lambda1_opt.BIC<-l.jgl$lambda1_l[which.min(BICval.JGL)]
  lambda2_opt.BIC<-l.jgl$lambda2_l[which.min(BICval.JGL)]
  
  # final estimation with optimal parameters
  jgl.BIC<- JGL.plug(S.list=S.list,n=n, 
                     lambda1=lambda1_opt.BIC, 
                     lambda2=lambda2_opt.BIC,
                     maxit=50,tol=1e-3)
  
  classes<-predict.DA(testY,means=mean_list, theta=jgl.BIC$theta, type="QDA",priors=n/sum(n))$cl

  return(list( estim=estim,mean_list=mean_list,S.list=S.list, 
               theta= jgl.BIC$theta, 
               lambda1_max=lambda1_max, lambda1_opt=lambda1_opt.BIC,
               lambda2_max=lambda2_max, lambda2_opt=lambda2_opt.BIC,
               predict = classes))
  
}
  