############################################
#
#  JGL.plug function
#  solve the joint graphical lasso from a list of K given covariance matrix estimates
#  instead of a data matrix.
#
#############################################
JGL.plug<- function (S.list, n, penalty = "fused", lambda1, lambda2, weights = "sample.size", 
                    penalize.diagonal = FALSE, 
                    maxit = 50, tol = 1e-03, warm = NULL, 
                    return.whole.theta = TRUE, screening = "fast", trunc = 1e-05) {
  
  # this function is an adaptation of the JGL function from the JGL package
  #########
  # INPUT #
  #########
  # S.list : list of estimated covariance matrices for JGL
  # penalty : penalty of JGL (possible values "fused" and "group")
  # lambda1 : regularization parameter on the elements of inverse error covariance matrix
  # lambda2 : regularization parameter on the difference between elements of inverse error covariance matrix (fused lasso)
  # rho : step size parameter : large value decrease step size
  # weights : putative sample size of each class'data . Allowed values : a vector of length K; 
  #           "equal" giving each class weight 1 ; 
  #           "sample.size", giving each class weight corresponding to its sample size 
  # penalize.diagonal :determines wether the penalties are applied to the diagonal elements
  # maxit : maximum number of iterations
  # tol : convergence criterion
  # warm :  input a warm start to theta in the form of a K-length list of pxp matrices
  # return.whole.theta : when TRUE, each class' inverse covariance matrix is returned
  # screening : "fast" recommended unless p prohibits storage of pxp matrices. 
  #             "memory.efficient" allow a solution with a much longer computation time for very high dimensional data
  # truncate : at convergence, all values below this threshold will be set to zero
  
  ##########
  # OUTPUT #
  ##########
  # theta : list of the estimated inverse covariance matrices
  
  ##########
  # Start #
  ##########
  p = dim(S.list[[1]])[2]
  K = length(S.list)
  rho<- sum(n)/K
  # step size parameter for the ADMM algorithm :
  # large value decrease step sizein Danaher paper, they takerho=1 but divide the objetcive function by n so this
  # is equivalent to taking rho=1
  
  ######################### 
  # Assign weights to each group
    if (length(weights) == 1 & weights == "equal") {
        weights = rep(1, K)
    } else if (length(weights) == 1 & weights == "sample.size") {
        weights = n
    }
  connected = rep(TRUE, p) # take all the features
 
  ######################### 
  # Computation of the covariance matrices
  # (only for the connected features)
  S<-vector("list",K)
  for (k in 1:K) {
    S[[k]] = S.list[[k]][connected , connected]
  }
  
  # if a penalty matrix is entered, only take its appropriate rows: 
  lam1 = lambda1
  lam2 = lambda2
  if (length(lam1) > 1) {
    lam1 = lam1[connected, connected]
  }
  if (length(lam2) > 1) {
    lam2 = lam2[connected, connected]
  }
  
  ######################### 
  # Determines wether the solution to the JGL optimization problem is block diagonal after permutation of the connected features
  # and identifies the blocks/partition/clusters
  # Then, we can simply perform the JGL algorithm on each block separately to obtain exactly
  # the same solution obtained on the whole dataset
  
  if (penalty == "fused") {
    
    #use bi-conditional screening rule to identify block structure exactly
    if (K == 2) {
      # criterion (a) and (b) in Danaher paper
      crit1 = list()
      for (k in 1:K) {
        crit1[[k]] = abs(S[[k]]) * weights[k] > lam1 + lam2 #  1 for connected features pairs
      }
      # criterion (c) in Danaher paper
      S.sum = matrix(0, sum(connected), sum(connected))
      for (k in 1:K) {
        S.sum = S.sum + weights[k] * S[[k]]
      }
      S.sum = abs(S.sum)
      crit2 = S.sum > 2 * lam1 # =1 for connected features pairs
    }
    
    # use sufficient screening rule to identify larger-grained block structure
    if (K > 2) {
      crit1 = list()
      for (k in 1:K) {
        crit1[[k]] = abs(S[[k]]) * weights[k] > lam1 # =1 for connected features pairs
      }
      crit2 = matrix(0, sum(connected), sum(connected))
    }
    
    # are both criteria met?
    critboth = crit2
    for (k in 1:K) {
      critboth = critboth + crit1[[k]] # !=0 for connected variables
    }
    critboth = (critboth != 0) 
    diag(critboth) = 1
      # pxp matrix with 1 for connected features pairs and 0 for unconnected features pairs
  }
  if (penalty == "group") {
    tempsum = matrix(0, sum(connected), sum(connected))
    for (k in 1:K) {
      tempsum = tempsum + (pmax(n[k] * abs(S[[k]]) - lam1, 0))^2
    }
    critboth = tempsum > lam2^2
    diag(critboth) = 1
  }
  
  # from this adjacency matrix, identify the partitions/clusters of features
  # in order to launch the JGL algorithm separately on each cluster/block 
    g1 <- graph.adjacency(critboth)
    cout = clusters(g1)  
      # cout$membership = numeric vector giving the cluster id to which each vertex belongs
      # cout$no = number of clusters
  
    blocklist = list() # will contain a list of vectors containing the feature numbers within each cluster
    unconnected = c()
    if (min(cout$membership) == 0) { # re-number the clusters starting with 1
      cout$membership = cout$membership + 1
    }
    
    for (i in 1:(cout$no)) {
    # when a cluster contains only one feature, add it to the unconnected features
    if (sum(cout$membership == i) == 1) {
      unconnected <- c(unconnected, which(cout$membership == i))
    }
    # when a cluster contains several features, update the list of blocks
    else if (sum(cout$membership == i) > 1) {
      blocklist[[length(blocklist) + 1]] <- which(cout$membership == i)
    }
  }
  # update the vector of connected features (0 for unconnected and 1 for connected)
  connected[unconnected] = FALSE 
  connected.index = rep(0, p) 
  connected.index[connected] = 1:sum(connected) 
      # vector of dimension p with F for unconnected features and numbers for connected features
  unconnected = !connected 
  
  
  
  
  ######################### 
  # Estimation of the inverse covariance matrices for connected features
  
    theta = list() # theta will contain a list of K inverse cov matrices with size equals to the number of connected features
    for (k in 1:K) {
      theta[[k]] = matrix(0, sum(connected), sum(connected))
      
    }
  
  # penalty vectors:
  # note: for admm.iters.unconnected, we use the penalize.diagonal argument before calling the function. 
  # for admm.iters, we use it IN the function.  
  
  if (length(lambda1) == 1) {
    lam1.unconnected = lambda1
  }else if (length(lambda1) > 1) {
    lam1.unconnected = diag(lambda1)[unconnected]
  }
  if (length(lambda2) == 1) {
    lam2.unconnected = lambda2
  }else if (length(lambda2) > 1) {
    lam2.unconnected = diag(lambda2)[unconnected]
  }
  
  # if penalize.diagonal==FALSE, then set the appropriate penalty vectors to zero:
  if (!penalize.diagonal) {
    lam1.unconnected = lam1.unconnected * 0
    if (penalty == "group") {
      lam2.unconnected = lam2.unconnected * 0
    }
  }
  
  ######################### 
  # ADMM algorithm for the unconnected features
  # get the unconnected portion of theta:
  
    Su = list()
    for (k in 1:K) {
      tmp<- S.list[[k]][unconnected, unconnected]
      if(is.matrix(tmp)){
        Su[[k]] = diag(S.list[[k]][unconnected, unconnected]) 
      }else {
        Su[[k]] = tmp
      }
    }
    
    if (sum(unconnected) > 0) {
    theta.unconnected = admm.iters.unconnected.plug(S.list= Su, lambda1 = lam1.unconnected, 
                                               lambda2 = lam2.unconnected, penalty = penalty, 
                                               rho = rho, rho.increment = 1,
                                               weights = weights, maxit = maxit, tol = tol)$Z

  }else if (sum(unconnected) == 0) {
    theta.unconnected = NULL
  }
  
  ######################### 
  # ADMM algorithm for connected features, for each block
  
  if (length(blocklist) > 0) {
    
    for (i in 1:length(blocklist)) {
      
      bl <- blocklist[[i]] 
      Sbl = list()
      for (k in 1:K) {
        Sbl[[k]] = S[[k]][bl , bl] # take only the features of the current block
      }
      if (length(lambda1) == 1) {
        lam1.bl = lambda1
      }else if (length(lambda1) > 1) {
        lam1.bl = lambda1[bl, bl]
      }
      if (length(lambda2) == 1){
        lam2.bl = lambda2
      }else if (length(lambda2) > 1) {
        lam2.bl = lambda2[bl, bl]
      }
      
      lam1.bl = penalty.as.matrix(lam1.bl, dim(Sbl[[1]])[2], penalize.diagonal = penalize.diagonal)
      if (penalty == "fused") {
        lam2.bl = penalty.as.matrix(lam2.bl, dim(Sbl[[1]])[2], penalize.diagonal = TRUE)
      }else if (penalty == "group") {
        lam2.bl = penalty.as.matrix(lam2.bl, dim(Sbl[[1]])[2], penalize.diagonal = penalize.diagonal)
      }
      if (length(warm) == 0) {
        warm.bl = NULL
      }
      else if (length(warm) > 0) {
        warm.bl = list()
        for (k in 1:K) {
          warm.bl[[k]] = warm[[k]][bl, bl] # take only the features of the current block in the warm start matrices
        }
      }
           
      # solve the JGL algorithm for the current block --> list of K matrices
      Thetabl = admm.iters.plug(S.list= Sbl, lambda1=lam1.bl, lambda2=lam2.bl,
                                penalty = penalty, rho.increment = 1,
                                rho = rho, weights = weights, penalize.diagonal = penalize.diagonal, 
                                maxit = maxit, tol = tol, warm = warm.bl)
      for (k in 1:K) {
        # in the final theta matrix, plug the estimated inverse covariance matrix for the current block
        theta[[k]][connected.index[bl], connected.index[bl]] = Thetabl$Z[[k]]
      }
    }
  }
  
  ######################### 
  # Elements below the threshold are truncated to zero
  if (dim(theta[[1]])[1] > 0) {
    for (k in 1:K) {
      rounddown = abs(theta[[k]]) < trunc
      diag(rounddown) = FALSE
      theta[[k]] = theta[[k]] * (1 - rounddown)
    }
  }
  if (!return.whole.theta) {
    out = list(theta = theta, theta.unconnected = theta.unconnected, 
               connected = connected)
  }
  
  ######################### 
  # Return the complete theta
  
  if (return.whole.theta) {
    whole.theta = list() # contains the list of k inverse covariance matrices (complete, i.e. with connected and unconnected features)
    for (k in 1:K) {
      whole.theta[[k]] = matrix(0, p, p)
      diag(whole.theta[[k]])[unconnected] = theta.unconnected[[k]] 
      # plug the estimated inverse variances on the diagonal for unconnected features
      whole.theta[[k]][connected, connected] = theta[[k]]
      # plug the estimated block  diagonal covariance matrices for the connected features
      #dimnames(whole.theta[[k]])[[1]] = dimnames(whole.theta[[k]])[[2]] = dimnames(Y[[k]])[[2]]
    }
    out = list(theta = whole.theta, connected = connected)
  }
  class(out) = "jgl"
  return(out)
}
