#############
# RDA
#############

RDA<-function(trainY, groupings, testY,
              estim="class",
                Maxmin.ratio=0.01, Maxmin.ratio2=0.01, nsteps=5,nsteps2=5){

    ######################
    # INPUTS
    ######################
    # trainY : Data matrix of dimension N X p (N: total sample size, p: number of variables)
    # groupings : numeric vector of dimension N containing the group membership of the observations
    # testY : Data matrix of dimension m X p  containing the m new observations to classify
    # estim : character string for the choice of location and covariance estimates
    #         either "class" for the sample estimators, or "pairwCorrK" for the cordinatewise median and the pairwise covariance matrices
    #         based on Kendall's correlation
    # Maxmin.ratio : ratio between the lower bound and the upper bound of the grid of values for the parameter rho1
    # Maxmin.ratio2 : ratio between the lower bound and the upper bound of the grid of values for the parameter rho2
    # nsteps: number of parameter values in the grid search for the optimal rho1
    # nsteps2: number of parameter values in the grid search for the optimal rho2
    
    ######################
    # OUTPUTS
    ######################
    # estim: choice of location and covariance estimates 
    # mean_list : list of mean estimates for each group
    # S.list : list of covariance matrix estimates for each group
    # S.RDA: list of regularized covariance matrix estimates for each group
    # rho1_max: maximum value of the grid of values for rho1
    # rho1_opt : optimal regularization parameter lambda1 minimizing the BIC
    # rho1_max: maximum value of the grid of values for rho2
    # rho1_opt : optimal regularization parameter lambda2 minimizing the BIC
    # classes: predicted groups for the observation of testY
    
    ######################
    # LIBRARIES
    ######################
    # library(huge)
    # library(pcaPP)
    # library(robustbase)
    
    ######################
    # AUXILIARY FUNCTIONS
    ######################
    # CellwiseCov.R
    # BICfct.R
    # predictDA.R
    
    ######################
    # ADDITIONAL FUNCTIONS
    ######################
    is.invertible <- function(m) class(try(solve(m),silent=T))=="matrix"
  
    RDAest<-function(S.list,S.pooled, rho_1, rho_2,n,p){
      temp1<-lapply(S.list,function(x, rho_1, S.pooled)(1-as.numeric(rho_1))*x + as.numeric(rho_1)*as.matrix(S.pooled), rho_1=rho_1, S.pooled=S.pooled)
      sigmaRDA<- lapply(temp1, function(x, rho_2,p)(1-rho_2)*x +rho_2/p *sum(diag(x))* diag(p),rho_2=rho_2, p=p)
      return(list(sigmaRDA=sigmaRDA, rho_1=rho_1, rho_2=rho_2))
    }
    
    BIC.RDA <-function( rho1,rho2, S.list,S.pooled, n, p){
      FIT<-RDAest(S.list=S.list, S.pooled, n=n,p=p,rho_1=rho1,rho_2=rho2) 
      if(Reduce('+',lapply(FIT$sigmaRDA,  is.invertible ))==length(S.list)){
        BIC_value<- BICfct(A=S.list, B= lapply(FIT$sigmaRDA,solve), n=n,p=p)
      }else{
        BIC_value=NA
      }
      return(BIC_value)
    }
    
  
  ##########
  # START
  ##########
  trainY<-split(data.frame(trainY),groupings)
  K<-length(unique(groupings))
  n<-unlist(lapply(trainY,nrow))
  p<-dim(trainY[[1]])[2]
  
  # Estimation of the means and covariance matrices
  if(estim=="class"){
    mean_list <- lapply(trainY, function(u)apply(u,2,mean))
    S.list<-lapply(trainY, cov)
  }else if(estim=="pairwCorrK"){
    mean_list<- lapply(trainY,function(t) apply(t,2,median))
    S.list<- lapply(trainY, kendall.transformed)
  }
  S.pooled<- Reduce('+',mapply(function(s,n)n*s,s=S.list,n=n,SIMPLIFY=F))/(sum(n)-K)
  
  # Grid of regularization parameters
  rho1_max<- 1 
  rho2_max<- 1
  rho1_seq<-sort(exp(seq(log(rho1_max),log(Maxmin.ratio*rho1_max),length.out=nsteps)),decreasing=T)
  rho2_seq<-sort(exp(seq(log(rho2_max),log(Maxmin.ratio2*rho2_max),length.out=nsteps2)),decreasing=T)
  r1<-list(rho1_l=sort(rep(rho1_seq,length(rho2_seq)),decreasing=T), rho2_l=rep(rho2_seq,length(rho1_seq)))
  
  # Selection of regularization parameters
  morearg.plug=list(S.list=S.list, S.pooled=S.pooled,n=n,p=p)
  BICval.RDA<-mapply(BIC.RDA,r1$rho1_l, r1$rho2_l, MoreArgs = morearg.plug,SIMPLIFY=F)
  rho1_opt.BIC<- r1$rho1_l[which.min(BICval.RDA)]
  rho2_opt.BIC<- r1$rho2_l[which.min(BICval.RDA)]

  # Final solution with optimal parameters
  RDA.BIC<- RDAest(S.list=S.list,S.pooled= S.pooled,
                rho_1=rho1_opt.BIC, rho_2=rho2_opt.BIC,
                n=n,p=p)
  classes<-predict.DA(testY, means=mean_list,theta=lapply(RDA.BIC$sigmaRDA, solve),type="QDA",priors=n/sum(n))$cl
  
  return(list( estim=estim,mean_list=mean_list,S.list=S.list, 
               S.RDA= RDA.BIC$sigmaRDA, 
               rho1_opt=rho1_opt.BIC,
               rho2_opt=rho2_opt.BIC,
               predict = classes))
}